@@ -5,6 +5,7 @@
 Usage:
   igloo [-bdmrs] [-p PROFILE | -u URL] FILENAME ...
   igloo [-abd] [-p PROFILE | -u URL] --list
+  igloo --add-url URL [PROFILE] | --delete-url PROFILE | --show-urls
   igloo -h | --help | --version
 
 Arguments:
@@ -35,7 +36,7 @@ Todo:
 
 """
 
-__version__ = '0.0.25'
+__version__ = '0.0.26'
 
 
 from codecs import getwriter
@@ -51,14 +52,14 @@ from traceback import format_exc
 try:
   from docopt import docopt
   from paramiko import SSHClient, SSHException
-  from yaml import load
+  from yaml import dump, load
 except ImportError:
   pass # probably in setup.py
 
 
 ERRORS = {
   0: 'something bad happened',
-  1: 'unable to connect to %r',
+  1: 'unable to connect to %r@%r',
   2: 'remote file %r not found',
   3: 'local file %r not found',
   4: 'transfer interrupted',
@@ -70,10 +71,40 @@ ERRORS = {
   10: 'profile %r not found in configuration file',
 }
 
-PATH = environ.get('IGLOO_CONFIG_PATH', expanduser(join('~', '.igloorc')))
 
+def get_stream_writer(binary=False, writer=stdout):
+  """Returns the stream writer used by the client."""
+  if binary:
+    return writer
+  else:
+    return getwriter(getpreferredencoding())(writer)
+
+def write(iterable, binary=False, lazy_flush=True, formatting='%s\n',
+  writer=stdout):
+  """Write to stdout, handles enconding automatically."""
+  writer = get_stream_writer(binary=binary, writer=writer)
+  for elem in iterable:
+    writer.write(formatting % elem)
+    if not lazy_flush:
+      writer.flush()
+  if lazy_flush:
+    writer.flush()
+
+def get_callback():
+  """Callback factory function for ``sftp.put`` and ``sftp.get``."""
+  writer = get_stream_writer()
+  def callback(transferred, total):
+    """Actual callback function."""
+    progress = int(100 * float(transferred) / total)
+    if progress < 100:
+      writer.write(' %2i%%\r' % (progress, ))
+    else:
+      writer.write('      \r')
+    writer.flush()
+  return callback
 
 def parse_url(url):
+  """Parse URL into user, host and remote directory."""
   if '@' in url:
     user, url = url.split('@', 1)
   else:
@@ -83,7 +114,11 @@ def parse_url(url):
   else:
     host = url
     path = '.'
-  return user, host, path
+  if not host:
+    raise ValueError('Empty url')
+  else:
+    return user, host, path
+
 
 class ClientError(Exception):
 
@@ -93,34 +128,62 @@ class ClientError(Exception):
 
   """
 
-  def __init__(self, number, details):
+  def __init__(self, number, details=()):
     super(ClientError, self).__init__('error: ' + ERRORS[number] % details)
-    self.number = number
-    self.details = details
     self.traceback = format_exc()
 
 
+class ClientOptions(object):
+
+  """Handles loading and saving of options (currently only profiles)."""
+
+  path = environ.get('MYIGLOORC', expanduser(join('~', '.igloorc')))
+
+  def __init__(self):
+    try:
+      with open(self.path) as handle:
+        self.config = load(handle)
+    except IOError:
+      self.config = {'profiles': {}}
+
+  def _save(self):
+    """Save options to file."""
+    with open(self.path, 'w') as handle:
+      dump(self.config, handle)
+
+  def _remove(self):
+    """Delete options files."""
+    remove(self.path)
+
+  def get_url(self, profile):
+    """Get URL corresponding to profile."""
+    try:
+      return self.config['profiles'][profile]
+    except KeyError:
+      raise ClientError(10, (profile, ))
+
+  def add_url(self, profile, url):
+    """Create new profile/URL entry."""
+    self.config['profiles'][profile] = url
+    self._save()
+
+  def delete_url(self, profile):
+    """Delete profile entry."""
+    try:
+      self.config.pop(profile)
+    except KeyError:
+      raise ClientError(10, (profile, ))
+    self._save()
+
+  def show_urls(self):
+    """Show all profile/URL entries."""
+    return self.config['profiles']
+
+
 class Client(object):
 
   """API client."""
 
-  @classmethod
-  def from_profile(cls, profile):
-    """Attempt to load configuration options.
-
-    :param profile: the profile to load
-    :rtype: :class:`Client`
-
-    """
-    try:
-      with open(PATH) as handle:
-        url = load(handle)[profile]
-    except IOError:
-      raise ClientError(9)
-    except KeyError:
-      raise ClientError(10, (profile, ))
-    return cls(url)
-
   def __init__(self, url, host_keys=None):
     self.user, self.host, self.path = parse_url(url)
     self.host_keys = host_keys or join(expanduser('~'), '.ssh', 'known_hosts')
@@ -154,31 +217,12 @@ class Client(object):
     finally:
       ssh.close()
 
-  def get_stream_writer(self, binary=False, writer=stdout):
-    """Returns the stream writer used by the client."""
-    if binary:
-      return writer
-    else:
-      return getwriter(getpreferredencoding())(writer)
-
-  def get_callback(self):
-    """Callback function for ``sftp.put`` and ``sftp.get``."""
-    writer = self.get_stream_writer()
-    def callback(transferred, total):
-      progress = int(100 * float(transferred) / total)
-      if progress < 100:
-        writer.write(' %2i%%\r' % (progress, ))
-      else:
-        writer.write('      \r')
-      writer.flush()
-    return callback
-
   def upload(self, filename, stream=False, track=False, move=False):
     """Attempt to upload a file the remote host."""
     with self.get_sftp_client() as sftp:
       if not stream:
         if track:
-          callback = self.get_callback()
+          callback = get_callback()
         else:
           callback = None
         try:
@@ -207,12 +251,12 @@ class Client(object):
       try:
         if not stream:
           if track:
-            callback = self.get_callback()
+            callback = get_callback()
           else:
             callback = None
           sftp.get(filename, filename, callback)
         else:
-          writer = self.get_stream_writer(binary=binary)
+          writer = get_stream_writer(binary=binary)
           remote_file = sftp.file(filename, 'rb')
           remote_file.prefetch()
           try:
@@ -234,30 +278,36 @@ class Client(object):
 
   def list(self, include_all=False):
     """Attempt to list available files on the remote host."""
-    writer = self.get_stream_writer()
     with self.get_sftp_client() as sftp:
-      filenames = (
+      return [
         filename
         for filename in sftp.listdir()
         if not filename.startswith(u'.') or include_all
-      )
-      writer.write('\n'.join(filenames))
-      writer.write('\n')
-      writer.flush()
+      ]
 
 
-def main():
-  """Command line parser. Docopt is amazing."""
-  arguments = docopt(__doc__, version=__version__)
-  try:
-    url = arguments['--url']
-    if url:
-      client = Client(url)
-    else:
-      client = Client.from_profile(arguments['--profile'])
+def parse_arguments(arguments):
+  """Main handler."""
+  options = ClientOptions()
+  if arguments['--add-url']:
+    options.add_url(
+      url=arguments['URL'],
+      profile=arguments['PROFILE'] or 'default',
+    )
+  elif arguments['--delete-url']:
+    options.delete_url(arguments['PROFILE'])
+  elif arguments['--show-urls']:
+    write(
+      sorted(reversed(options.show_urls().items())),
+      formatting='%s [%s]\n'
+    )
+  else:
+    url = arguments['--url'] or options.get_url(arguments['--profile'])
+    client = Client(url)
     if arguments['--list']:
-      client.list(
-        include_all=arguments['--all'],
+      write(
+        client.list(include_all=arguments['--all']),
+        binary=arguments['--binary'],
       )
     elif arguments['--remote']:
       for filename in arguments['FILENAME']:
@@ -274,6 +324,13 @@ def main():
           stream=arguments['--stream'],
           move=arguments['--move'],
         )
+
+
+def main():
+  """Command line parser. Docopt is amazing."""
+  arguments = docopt(__doc__, version=__version__)
+  try:
+    parse_arguments(arguments)
   except ClientError as err:
     if arguments['--debug']:
       stderr.write(err.traceback)
